#include "Bloc.h"
#include "Sommet.h"
#include "Util.h"
#include <algorithm>


const WORD
   Bloc::IndexBloc [NB_SOMMETS_PRIMITIVES] =
   {
      0, 1, 2,     // devant
      0, 2, 3,     // devant
      4, 5, 7,     // arrire
      5, 6, 7,     // arrire
      8, 9, 11,    // dessous
      9, 10, 11,   // dessous
      12, 13, 15,  // dessus
      13, 14, 15,  // dessus
      16, 17, 18,  // gauche
      18, 19, 16,  // gauche
      21, 22, 23,  // droite
      21, 23, 20   // droite
   };

Bloc::Bloc (D3DXVECTOR3 _origine, D3DXVECTOR3 _dimension,
            IDirect3DDevice9 *pDevice)
   : IDessinable (_origine + _dimension / 2),
     m_pIB (0), m_pVB (0), m_Materiau ()
{
   SOMMET *Sommets;
   pDevice->CreateVertexBuffer ( NB_SOMMETS * sizeof (SOMMET),
                                 0,
                                 SOMMET::D3DFVF_SOMMET,
                                 D3DPOOL_DEFAULT,
                                 &m_pVB,
                                 0);
   // Bloquer la mmoire systme et obtenir un pointeur
   // sur le vertex buffer
   m_pVB->Lock (0, 0, reinterpret_cast<void **>(&Sommets), 0);

   // Calculer le coin le plus loign du bloc
   D3DXVECTOR3 extent =  _origine + _dimension;

   // Calculer les normales
   enum Faces { Devant, Arriere, Dessous, Dessus, Gauche, Droite };
   D3DXVECTOR3 Normales [NB_FACES] =
   {
      D3DXVECTOR3 ( 0.0f,  0.0f, -1.0f), // devant
      D3DXVECTOR3 ( 0.0f,  0.0f,  1.0f), // arrire
      D3DXVECTOR3 ( 0.0f, -1.0f,  0.0f), // dessous
      D3DXVECTOR3 ( 0.0f,  1.0f,  0.0f), // dessus
      D3DXVECTOR3 (-1.0f,  0.0f,  0.0f), // face gauche
      D3DXVECTOR3 ( 1.0f,  0.0f,  0.0f)  // face droite
   };

   // -- Sommets du bloc
   // Points
   D3DXVECTOR3 Pt[NB_SOMMETS_DISTINCTS] =
   {
      D3DXVECTOR3 (_origine.x, _origine.y, _origine.z),
      D3DXVECTOR3 (_origine.x, extent.y, _origine.z),
      D3DXVECTOR3 (extent.x, extent.y, _origine.z),
      D3DXVECTOR3 (extent.x, _origine.y, _origine.z),
      D3DXVECTOR3 (extent.x, _origine.y, extent.z),
      D3DXVECTOR3 (extent.x, extent.y, extent.z),
      D3DXVECTOR3 (_origine.x, extent.y, extent.z),
      D3DXVECTOR3 (_origine.x, _origine.y, extent.z)
   };

   // Devant du bloc
   Sommets[0]  = SOMMET (Pt[0], Normales[0]);
   Sommets[1]  = SOMMET (Pt[1], Normales[0]);
   Sommets[2]  = SOMMET (Pt[2], Normales[0]);
   Sommets[3]  = SOMMET (Pt[3], Normales[0]);

   // Arrire du bloc
   Sommets[4]  = SOMMET (Pt[4], Normales[1]);
   Sommets[5]  = SOMMET (Pt[5], Normales[1]);
   Sommets[6]  = SOMMET (Pt[6], Normales[1]);
   Sommets[7]  = SOMMET (Pt[7], Normales[1]);

   // Dessous du bloc
   Sommets[8]  = SOMMET (Pt[0], Normales[2]);
   Sommets[9]  = SOMMET (Pt[3], Normales[2]);
   Sommets[10] = SOMMET (Pt[4], Normales[2]);
   Sommets[11] = SOMMET (Pt[7], Normales[2]);

   // Dessus du bloc
   Sommets[12] = SOMMET (Pt[1], Normales[3]);
   Sommets[13] = SOMMET (Pt[6], Normales[3]);
   Sommets[14] = SOMMET (Pt[5], Normales[3]);
   Sommets[15] = SOMMET (Pt[2], Normales[3]);

   // Face gauche du bloc
   Sommets[16] = SOMMET (Pt[0], Normales[4]);
   Sommets[17] = SOMMET (Pt[7], Normales[4]);
   Sommets[18] = SOMMET (Pt[6], Normales[4]);
   Sommets[19] = SOMMET (Pt[1], Normales[4]);

   // Face droite du bloc
   Sommets[20] = SOMMET (Pt[3], Normales[5]);
   Sommets[21] = SOMMET (Pt[2], Normales[5]);
   Sommets[22] = SOMMET (Pt[5], Normales[5]);
   Sommets[23] = SOMMET (Pt[4], Normales[5]);

   // Librer le Vertex Buffer
   m_pVB -> Unlock ();

   // Initialsiation de l'index buffer
   InitIndexBuffer (pDevice);

   // Initialisation des matriaux
   InitMateriaux ();
}

void Bloc::InitIndexBuffer (IDirect3DDevice9 *pDevice)
{
   WORD *pIndex;
   // Cration du buffer d'index
   pDevice ->CreateIndexBuffer(NB_SOMMETS_PRIMITIVES * sizeof (WORD),
                               D3DUSAGE_WRITEONLY,
                               D3DFMT_INDEX16,
                               D3DPOOL_DEFAULT,
                               &m_pIB,
                               0);

   // Remplir le buffer d'index
   m_pIB->Lock (0, 0, reinterpret_cast<void **> (&pIndex), 0);
   std::copy (&IndexBloc[0], &IndexBloc[NB_SOMMETS_PRIMITIVES], pIndex);
   m_pIB->Unlock ();
}

void Bloc::Draw (IDirect3DDevice9 *pDevice)
{
   // Matrice de transformation dans le monde
   pDevice->SetTransform (D3DTS_WORLD, &m_MatriceMonde);
   DrawNoTransform (pDevice);
}

void Bloc::DrawNoTransform (IDirect3DDevice9 *pDevice)
{
   // Choisir le matriau
   pDevice->SetMaterial (&m_Materiau);
   // Source des sommets
   pDevice->SetStreamSource (0, m_pVB, 0, sizeof (SOMMET));
   // Type des sommets
   pDevice->SetFVF (SOMMET::D3DFVF_SOMMET);
   // Sources des index
   pDevice->SetIndices (m_pIB);
   // Affichage
   pDevice->DrawIndexedPrimitive (D3DPT_TRIANGLELIST, 0, 0, NB_SOMMETS, 0, NB_PRIMITIVES);
}

void Bloc::SetMateriaux (const D3DMATERIAL9 &Materiau)
{
   m_Materiau = Materiau;
}

void Bloc::InitMateriaux ()
{
   ::memset (&m_Materiau, 0, sizeof (D3DMATERIAL9));
   // Initialiser le RGBA pour la rflexion ambiante
   m_Materiau.Ambient.r = 0.5f;
   m_Materiau.Ambient.g = 0.0f;
   m_Materiau.Ambient.b = 0.5f;
   m_Materiau.Ambient.a = 1.0f;
}

Bloc::~Bloc ()
{
   Relacher (m_pVB);
   Relacher (m_pIB);
}
