#include "Plan.h"
#include "Sommet.h"
#include "Util.h"
#include <algorithm>







class MorceauDePlan : public IDessinable
{
   IDirect3DVertexBuffer9 *m_pVB; // vertex buffer du plan
   IDirect3DIndexBuffer9 *m_pIB; // index buffer du plan
   D3DMATERIAL9 m_Materiau; // matriau du plan
   void InitIndexBuffer (IDirect3DDevice9 *pDevice);
   enum
   {
      NB_SOMMETS = 4,
      NB_FACES = 2,
      NB_SOMMETS_DISTINCTS = NB_SOMMETS * NB_FACES,
      NB_PRIMITIVES_FACE = 2,
      NB_SOMMETS_PRIMITIVE = 3,
      NB_PRIMITIVES = NB_PRIMITIVES_FACE * NB_FACES
   };
public:
   static const float
      HAUTEUR, LARGEUR;
   MorceauDePlan(const D3DXVECTOR3 &_origine,
                 IDirect3DDevice9 *pDevice);
   virtual ~MorceauDePlan();

   void InitMateriaux ();
   virtual void Draw(IDirect3DDevice9 *pDevice);
   virtual void DrawNoTransform(IDirect3DDevice9 *pDevice);
};

const float
   MorceauDePlan::HAUTEUR = 5.0f,
   MorceauDePlan::LARGEUR = 5.0f;

MorceauDePlan::MorceauDePlan(const D3DXVECTOR3 &_origine, IDirect3DDevice9 *pDevice)
   : IDessinable (_origine),
     m_pIB (0), m_pVB (0), m_Materiau ()
{
   SOMMET *Sommets;
   pDevice->CreateVertexBuffer ( NB_SOMMETS_DISTINCTS * sizeof (SOMMET),
                                 0,
                                 SOMMET::D3DFVF_SOMMET,
                                 D3DPOOL_DEFAULT,
                                 &m_pVB,
                                 0);
   m_pVB->Lock (0, 0, reinterpret_cast<void **>(&Sommets), 0);
   D3DXVECTOR3 Normale = D3DXVECTOR3 ( 0.0f,  0.0f,  1.0f ); // dessus
   Sommets[0] = SOMMET (D3DXVECTOR3 (-LARGEUR/2, -HAUTEUR/2, 0.0f) + _origine,  Normale);
   Sommets[1] = SOMMET (D3DXVECTOR3 (-LARGEUR/2,  HAUTEUR/2, 0.0f) + _origine,  Normale);
   Sommets[2] = SOMMET (D3DXVECTOR3 ( LARGEUR/2,  HAUTEUR/2, 0.0f) + _origine,  Normale);
   Sommets[3] = SOMMET (D3DXVECTOR3 ( LARGEUR/2, -HAUTEUR/2, 0.0f) + _origine,  Normale);
   Sommets[4] = SOMMET (D3DXVECTOR3 (-LARGEUR/2, -HAUTEUR/2, 0.0f) + _origine, -Normale);
   Sommets[5] = SOMMET (D3DXVECTOR3 (-LARGEUR/2,  HAUTEUR/2, 0.0f) + _origine, -Normale);
   Sommets[6] = SOMMET (D3DXVECTOR3 ( LARGEUR/2,  HAUTEUR/2, 0.0f) + _origine, -Normale);
   Sommets[7] = SOMMET (D3DXVECTOR3 ( LARGEUR/2, -HAUTEUR/2, 0.0f) + _origine, -Normale);
   m_pVB -> Unlock ();
   InitIndexBuffer (pDevice);
   InitMateriaux ();
}

void MorceauDePlan::InitIndexBuffer (IDirect3DDevice9 *pDevice)
{
   const int NB_INDEX = NB_FACES * NB_PRIMITIVES_FACE * NB_SOMMETS_PRIMITIVE;
   const WORD SourceIndex[NB_INDEX] =
   {
      0, 1, 2,
      0, 2, 3,
      2, 1, 0,
      3, 2, 0
   };
   WORD *pIndex;
   pDevice ->CreateIndexBuffer(NB_INDEX * sizeof (WORD),
                               D3DUSAGE_WRITEONLY,
                               D3DFMT_INDEX16,
                               D3DPOOL_DEFAULT,
                               &m_pIB,
                               0);
   m_pIB->Lock (0, 0, reinterpret_cast<void **> (&pIndex), 0);
   std::copy (&SourceIndex[0], &SourceIndex[NB_INDEX], pIndex);
   m_pIB->Unlock ();
}

void MorceauDePlan::Draw (IDirect3DDevice9 *pDevice)
{
   // Matrice de transformation dans le monde
   pDevice->SetTransform (D3DTS_WORLD, &m_MatriceMonde);
   DrawNoTransform (pDevice);
}

void MorceauDePlan::DrawNoTransform (IDirect3DDevice9 *pDevice)
{
   const int NB_INDEX = NB_FACES * NB_PRIMITIVES_FACE * NB_SOMMETS_PRIMITIVE;
   pDevice->SetMaterial (&m_Materiau);
   pDevice->SetStreamSource (0, m_pVB, 0, sizeof (SOMMET));
   pDevice->SetFVF (SOMMET::D3DFVF_SOMMET);
   pDevice->SetIndices (m_pIB);
   pDevice->DrawIndexedPrimitive (D3DPT_TRIANGLELIST, 0, 0, NB_INDEX, 0, NB_PRIMITIVES);
}

void MorceauDePlan::InitMateriaux ()
{
   ::memset (&m_Materiau, 0, sizeof (D3DMATERIAL9));
   // Initialiser le RGBA pour la rflection ambiante
   m_Materiau.Ambient.r = 0.25f;
   m_Materiau.Ambient.g = 0.75f;
   m_Materiau.Ambient.b = 0.25f;
   m_Materiau.Ambient.a = 1.0f;
}

MorceauDePlan::~MorceauDePlan()
{
   Relacher (m_pVB);
   Relacher (m_pIB);
}








const int
   Plan::NB_CASES_HAUT_DEFAUT = 1,
   Plan::NB_CASES_LARGE_DEFAUT = 1;



Plan::Plan(const D3DXVECTOR3 &_origine, IDirect3DDevice9 *pDevice,
           const int NbCasesHaut, const int NbCasesLarge)
   : IDessinable (_origine),
     m_NbCasesHaut (NbCasesHaut), m_NbCasesLarge (NbCasesLarge)
{
   for (int i = 0; i < m_NbCasesHaut; i++)
      for (int j = 0; j < m_NbCasesLarge; j++)
      {
         D3DXVECTOR3 Pos (_origine.x + j * MorceauDePlan::LARGEUR,
                          _origine.y + i * MorceauDePlan::HAUTEUR,
                          _origine.z);
         m_Cases.push_back (new MorceauDePlan (Pos, pDevice));
      }
}

void Plan::Draw (IDirect3DDevice9 *pDevice)
{
   // Matrice de transformation dans le monde
   pDevice->SetTransform (D3DTS_WORLD, &m_MatriceMonde);
   DrawNoTransform (pDevice);
}

namespace
{
   template <class T> void Detruire (T p) { delete p; }
   class Dessiner
   {
      IDirect3DDevice9 *m_pDevice;
   public:
      Dessiner (IDirect3DDevice9 *pDevice) : m_pDevice (pDevice) { }
      virtual ~Dessiner () {}
      void operator () (IDessinable *p) { p->DrawNoTransform (m_pDevice); }
   };
}

void Plan::DrawNoTransform (IDirect3DDevice9 *pDevice)
{
   std::for_each (m_Cases.begin (), m_Cases.end (), ::Dessiner (pDevice));
}

Plan::~Plan ()
{
   std::for_each (m_Cases.begin (), m_Cases.end (), ::Detruire<MorceauDePlan *>);
}
