#include "Pyramide.h"
#include "Sommet.h"
#include "Util.h"
#include <algorithm>
#include <cmath>
#include <fstream>
#include <iostream>

const WORD
   Pyramide::Index [NB_SOMMETS_PRIMITIVES] =
   {
      0, 1, 2,     // base
      0, 2, 3,     // base
      4, 0, 3,
      2, 4, 3,
      1, 4, 2,
      0, 4, 1
   };

class NormOut
{
   std::ofstream *ofs;
public:
   NormOut () : ofs (new std::ofstream ("C:\\Mr_Roy\\Normales.txt")) {}
   ~NormOut () { if (ofs) delete ofs; }
   std::ostream &GetFlot () { return *ofs; }
   template <class T> std::ostream & operator << (const T &v) { *ofs << v; return *ofs; }
};
namespace { NormOut no; }

// Normale de v1v2^v2v3
D3DXVECTOR3 CalculerNormale (const D3DXVECTOR3 &v1,
                             const D3DXVECTOR3 &v2,
                             const D3DXVECTOR3 &v3)
{
   // Normale  la face AB^AC
   //    (yB - yA) * (zC - zA) - (yC - yA) * (zB - zA),
   //    (zB - zA) * (xC - xA) - (zC - zA) * (xB - xA),
   //    (xB - xA) * (yC - yA) - (xC - xA) * (yB - yA)
   D3DXVECTOR3 Resultat ( (v2.y-v1.y) * (v3.z-v1.z) - (v3.y-v1.y) * (v2.z-v1.z),
                          (v2.z-v1.z) * (v3.x-v1.x) - (v3.z-v1.z) * (v2.y-v1.y),
                          (v2.x-v1.x) * (v3.y-v1.y) - (v3.x-v1.x) * (v2.y-v1.y));
   ::no.GetFlot () << "Normale: " << Resultat.x << "," << Resultat.y << "," << Resultat.z << std::endl;
   return Resultat;
}

Pyramide::Pyramide (const D3DXVECTOR3 &Position,
                    const float LgCote,
                    IDirect3DDevice9 *pDevice)
   : IDessinable (Position), // centre de la base
     m_pIB (0), m_pVB (0), m_Materiau ()
{
   SOMMET *Sommets;
   pDevice->CreateVertexBuffer ( NB_SOMMETS * sizeof (SOMMET),
                                 0,
                                 SOMMET::D3DFVF_SOMMET,
                                 D3DPOOL_DEFAULT,
                                 &m_pVB,
                                 0);
   m_pVB->Lock (0, 0, reinterpret_cast<void **>(&Sommets), 0);
   const float LgDiagonaleBase = ::sqrt (2 * ::pow (LgCote, 2));
   const float HAUTEUR = LgCote;

   // Points
   D3DXVECTOR3 Pt[NB_SOMMETS_DISTINCTS] =
   {
      D3DXVECTOR3 (-LgDiagonaleBase/2, -LgDiagonaleBase/2, 0) + Position,
      D3DXVECTOR3 (-LgDiagonaleBase/2,  LgDiagonaleBase/2, 0) + Position,
      D3DXVECTOR3 ( LgDiagonaleBase/2,  LgDiagonaleBase/2, 0) + Position,
      D3DXVECTOR3 ( LgDiagonaleBase/2, -LgDiagonaleBase/2, 0) + Position,
      D3DXVECTOR3 ( 0, 0, HAUTEUR) + Position
   };

   D3DXVECTOR3 Normales [NB_FACES] =
   {
      D3DXVECTOR3 ( 0.0f,  0.0f, -1.0f), // base
      CalculerNormale (D3DXVECTOR3 (Pt[4]-Pt[0]), D3DXVECTOR3 (Pt[0]-Pt[3]), D3DXVECTOR3 (Pt[3]-Pt[4])),
      CalculerNormale (D3DXVECTOR3 (Pt[2]-Pt[4]), D3DXVECTOR3 (Pt[3]-Pt[2]), D3DXVECTOR3 (Pt[4]-Pt[3])),
      CalculerNormale (D3DXVECTOR3 (Pt[4]-Pt[1]), D3DXVECTOR3 (Pt[2]-Pt[4]), D3DXVECTOR3 (Pt[1]-Pt[2])),
      CalculerNormale (D3DXVECTOR3 (Pt[1]-Pt[0]), D3DXVECTOR3 (Pt[4]-Pt[1]), D3DXVECTOR3 (Pt[0]-Pt[4]))
   };

   // Base
   Sommets[0]  = SOMMET (Pt[0], Normales[0]);
   Sommets[1]  = SOMMET (Pt[1], Normales[0]);
   Sommets[2]  = SOMMET (Pt[2], Normales[0]);
   Sommets[3]  = SOMMET (Pt[3], Normales[0]);

   // Cts
   Sommets[4]  = SOMMET (Pt[4], Normales[1]);
   Sommets[5]  = SOMMET (Pt[0], Normales[1]);
   Sommets[6]  = SOMMET (Pt[3], Normales[1]);

   Sommets[7]  = SOMMET (Pt[2], Normales[2]);
   Sommets[8]  = SOMMET (Pt[4], Normales[2]);
   Sommets[9]  = SOMMET (Pt[3], Normales[2]);

   Sommets[10] = SOMMET (Pt[1], Normales[3]);
   Sommets[11] = SOMMET (Pt[4], Normales[3]);
   Sommets[12] = SOMMET (Pt[2], Normales[3]);

   Sommets[13] = SOMMET (Pt[0], Normales[4]);
   Sommets[14] = SOMMET (Pt[4], Normales[4]);
   Sommets[15] = SOMMET (Pt[1], Normales[4]);

   m_pVB -> Unlock ();

   // Initialsiation de l'index buffer
   InitIndexBuffer (pDevice);

   // Initialisation des matriaux
   InitMateriaux ();
}

void Pyramide::InitIndexBuffer (IDirect3DDevice9 *pDevice)
{
   WORD *pIndex;
   // Cration du buffer d'index
   pDevice ->CreateIndexBuffer(NB_SOMMETS_PRIMITIVES * sizeof (WORD),
                               D3DUSAGE_WRITEONLY,
                               D3DFMT_INDEX16,
                               D3DPOOL_DEFAULT,
                               &m_pIB,
                               0);

   // Remplir le buffer d'index
   m_pIB->Lock (0, 0, reinterpret_cast<void **> (&pIndex), 0);
   std::copy (&Index[0], &Index[NB_SOMMETS_PRIMITIVES], pIndex);
   m_pIB->Unlock ();
}

void Pyramide::Draw (IDirect3DDevice9 *pDevice)
{
   // -- Rendre l'objet
   // Matrice de transformation dans le monde
   pDevice->SetTransform (D3DTS_WORLD, &m_MatriceMonde);
   DrawNoTransform (pDevice);
}

void Pyramide::DrawNoTransform (IDirect3DDevice9 *pDevice)
{
   // Choisir le matriau
   pDevice->SetMaterial (&m_Materiau);
   // Source des sommets
   pDevice->SetStreamSource (0, m_pVB, 0, sizeof (SOMMET));
   // Type des sommets
   pDevice->SetFVF (SOMMET::D3DFVF_SOMMET);
   // Sources des index
   pDevice->SetIndices (m_pIB);
   // Affichage
   pDevice->DrawIndexedPrimitive (D3DPT_TRIANGLELIST, 0, 0, NB_SOMMETS, 0, NB_PRIMITIVES);
}

void Pyramide::SetMateriaux (const D3DMATERIAL9 &Materiau)
{
   m_Materiau = Materiau;
}

void Pyramide::InitMateriaux ()
{
   ::memset (&m_Materiau, 0, sizeof (D3DMATERIAL9));
   // Initialiser le RGBA pour la rflexion ambiante
   m_Materiau.Ambient.r = 1.0f;
   m_Materiau.Ambient.g = 0.0f;
   m_Materiau.Ambient.b = 0.0f;
   m_Materiau.Ambient.a = 1.0f;
}

Pyramide::~Pyramide ()
{
   Relacher (m_pVB);
   Relacher (m_pIB);
}
